import es
import pprint

info = es.AddonInfo()
info.name = "Hax Tools"
info.version = "1.2"
info.author = "GODJonez"
info.description = "Advanced console commands for debugging Python scripts"
info.basename = "haxtools"
info.url = "http://addons.eventscripts.com/addons/view/haxtools"


imports = {}
vfuncs = {}

def load():
    if not es.exists('command', 'i_script'):
        es.regcmd('i_script', 'haxtools/cmd_i_script', 'Add script to be tracked')
        es.regcmd('i_lib', 'haxtools/cmd_i_lib', 'Add library to be tracked')
        es.regcmd('i_reload', 'haxtools/cmd_i_reload', 'Force reload of module')
        es.regcmd('i_stop', 'haxtools/cmd_i_stop', 'Stop tracking')
        es.regcmd('i_hook', 'haxtools/cmd_i_hook', 'Block a function/method')
        es.regcmd('i_info', 'haxtools/cmd_i_info', 'Show information about module')
    es.dbgmsg(0, 'Python hax tools loaded. Available commands:')
    es.dbgmsg(0, ' i_script, i_lib, i_stop')
    es.dbgmsg(0, ' i_reload, i_hook, i_info')

def unload():
    for key in imports:
        RemoveInspections(key)
    es.dbgmsg(0, 'Python script inspector unloaded.')

class HackModule(object):
    def __init__(self):
        self.classes = {}
        self.vfuncs = {}
    def __getitem__(self, x):
        return self.vfuncs[x]
    def __setitem__(self, x, y):
        self.vfuncs[x] = y
    def __delitem__(self, x):
        del self.vfuncs[x]
    def __iter__(self):
        self.iter = self.vfuncs.iterkeys()
        return self
    def next(self):
        try:
            return self.iter.next()
        except StopIteration:
            raise StopIteration

def AddInspections(key, mod):
    objs = mod.__dict__
    vfuncs[key] = HackModule()
    RecursiveInspection(key, vfuncs[key], objs)

def RecursiveInspection(parent, vfd, objs):
    es.dbgmsg(1, 'RECURSIVEINSPECTION')
    nx = None
    for obj in objs.keys():
        if callable(objs[obj]):
            if type(objs[obj]).__name__ == 'function':
                es.dbgmsg(1, 'Inspector: found %s.%s'%(parent,obj))
                x = objs[obj]
#                print "x=",str(x)
#                print "objs[obj]=",str(objs[obj])
#                print "nx=",str(nx)
                nx = objs[obj] = GimmeInspectorLambda(parent, x)
                vfd[obj] = (x,nx)
            elif type(objs[obj]).__name__ == 'classobj':
                objs2 = objs[obj].__dict__
                vfd2 = vfd.classes[obj] = HackModule()
                RecursiveInspection("%s.%s"%(parent,obj), vfd2, objs2)

def RemoveInspections(key):
    mod = imports[key]
    objs = mod.__dict__
    RecursiveRemoving(key, vfuncs[key], objs)
    del vfuncs[key]
    
def RecursiveRemoving(parent, vfd, objs):
    for obj in vfd:
        es.dbgmsg(1, 'Inspector: removed %s.%s'%(parent,obj))
        objs[obj] = vfd[obj][0]
    for c in vfd.classes:
        es.dbgmsg(1, 'Inspector removing class %s'%c)
        vfd2 = vfd.classes[c]
        objs2 = objs[c].__dict__
        RecursiveRemoving("%s.%s"%(parent,c), vfd2, objs2)


def GimmeInspectorLambda(key, x):
    return lambda *p, **k: InspectFunc(key, x, *p, **k)
def GimmeHookerLambda(key):
    return lambda *p, **k: HookedFunction(key, *p, **k)



def InspectFunc(module, function, *params, **kw):
    es.dbgmsg(0, 'Inspector: %s.%s%s%s'%(
        module, function.__name__, str(params), str(kw)))
    function(*params,**kw)


def HookedFunction(function, *params, **kw):
    es.dbgmsg(0, 'Blocked: %s%s%s'%(function, str(params), str(kw)))


def cmd_i_script():
    if es.getargc() == 1:
        es.dbgmsg(0, 'i_script <basename>')
        es.dbgmsg(0, '  - Add script to be tracked')
    else:
        param = es.getargv(1)
        sn = str(param)
        try:
            imp = __import__("%s.%s"%(sn,sn))
            mod = imports[param] = imp.__dict__[sn]
            AddInspections(param, mod)
        except (KeyError, ImportError):
            es.dbgmsg(0, 'Script not found')


def cmd_i_lib():
    if es.getargc() == 1:
        es.dbgmsg(0, 'i_lib <name>')
        es.dbgmsg(0, '  - Add library to be tracked')
    else:
        param = es.getargv(1)
        sn = str(param)
        try:
            mod = imports[param] = __import__(sn)
            AddInspections(param, mod)
        except ImportError:
            es.dbgmsg(0, 'Module not found')


def cmd_i_reload():
    if es.getargc() == 1:
        es.dbgmsg(0, 'i_reload <name>')
        es.dbgmsg(0, '  - Reload library or script')
    else:
        param = es.getargv(1)
        if imports.has_key(param):
            RemoveInspections(param)
            mod = imports[param]
            mod = imports[param] = reload(mod)
            AddInspections(param, mod)
        else:
            try:
                reload(__import__(param).__dict__[param])
            except (ImportError, KeyError):
                try:
                    reload(__import__(param))
                except ImportError:
                    es.dbgmsg(0, 'Module not found')


def cmd_i_stop():
    if es.getargc() == 1:
        es.dbgmsg(0, 'i_stop <name>')
        es.dbgmsg(0, '  - Stop tracking a library or script')
    else:
        param = es.getargv(1)
        if imports.has_key(param):
            RemoveInspections(param)
            del imports[param]
        else:
            es.dbgmsg(0, 'Module was not tracked')

def cmd_i_hook():
    if es.getargc() == 1:
        es.dbgmsg(0, 'i_hook <module>.[<class>.]<function>')
        es.dbgmsg(0, '  - Block selected function')
    else:
        param = es.getargv(1)
        params = param.split('.')
        d = imports
        try:
            maxi = len(params)-1
            for i in range(maxi):
                d = d[params[i]].__dict__
            df = d[params[maxi]]
            if type(df).__name__ == 'function':
                es.dbgmsg(1, 'Hooking %s'%param)
                d[params[maxi]] = GimmeHookerLambda(param)
        except (AttributeError, KeyError, TypeError):
            es.dbgmsg(0, "Couldn't hook. Make sure the module is being inspected.")

def cmd_i_info():
    argc = es.getargc()
    if argc == 1:
        es.dbgmsg(0, 'i_info <module> [level]')
        es.dbgmsg(0, '  - Show information about a module')
    else:
        param = es.getargv(1)
        try:
            level = int(es.getargv(2))
        except ValueError:
            level = 1
        sn = str(param)
        try:
            imp = __import__(sn)
        except (ImportError), e:
            es.dbgmsg(0, 'Error: '+str(e))
            return False
        if sn in imp.__dict__ and level == 1:
            level = 2
        _showinfo(imp, level, "  ")

def _showinfo(imp, level, indent):
    classes = []
    functions = []
    variables = []
    modules = []
    
    d = imp.__dict__
    
    for item in d.items():
        if item[0].startswith('__'):
            continue
        t = type(item[1]).__name__
        if t in ('classobj', 'type'):
            classes.append(item)
        elif t in ('function','builtin_function_or_method','instancemethod'):
            functions.append(item)
        elif t in ('module',):
            modules.append(item)
        else:
            variables.append(item)

    es.dbgmsg(0, '%sObjects in "%s":'%(indent, imp.__name__))
    if variables:
        es.dbgmsg(0, '%s %d variables:'%(indent,len(variables)))
        variables.sort()
        for item in variables:
            pretty_print(indent+"  ", item[0], item[1], type(item[1]))
    else:
        es.dbgmsg(0, '%s No variables'%indent)
        
    if functions:
        es.dbgmsg(0, '%s %d functions:'%(indent,len(functions)))
        functions.sort()
        for item in functions:
            es.dbgmsg(0, '%s  %s'%(indent, item[0]))
    else:
        es.dbgmsg(0, '%s No functions'%indent)
        
    if classes:
        es.dbgmsg(0, '%s %d classes:'%(indent,len(classes)))
        classes.sort()
        if level > 1:
            for item in classes:
                _showinfo(item[1], level-1, indent+"  ")
        else:
            for item in classes:
                es.dbgmsg(0, '%s  %s'%(indent, item[0]))
    else:
        es.dbgmsg(0, '%s No classes'%indent)
    
    if modules:
        es.dbgmsg(0, '%s %d modules:'%(indent,len(modules)))
        modules.sort()
        if level > 1:
            for item in modules:
                _showinfo(item[1], level-1, indent+"  ")
        else:
            for item in modules:
                es.dbgmsg(0, '%s  %s'%(indent, item[0]))
    else:
        es.dbgmsg(0, '%s No modules'%indent)


def pretty_print(indent, name, text, typetext):
    fmt = pprint.pformat(text, 2).splitlines()
    if len(fmt) > 1:
        es.dbgmsg(0, '%s%s:   %s'%(indent, name, typetext))
        for line in fmt:
            es.dbgmsg(0, indent+" "+line)
    else:
        es.dbgmsg(0, '%s%s=%s %s'%(indent, name, repr(text), typetext))
